#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "xmalloc.h"
#include "lp.h"
#include "tableau.h"


void
free_tableau (struct TABLEAU *tp)
{
  int i;
  free (tp->chk);
  free (tp->imin);
  free (tp->jmax);
  free (tp->neg_costcoef_indices );
  free (tp->neg_rhs);
  free (tp->thmin);
  free (tp->delmax);
  for(i = 0; i < tp->maxrows; i++)
    free (tp->elements[i]);

  free (tp->elements);
}

struct TABLEAU *
new_tableau(void)
{
  struct TABLEAU *tp= xcalloc (1, sizeof (struct TABLEAU));

return tp;
}

void
tableau_set_rowscolumns (struct TABLEAU *t, int r, int c)
{
  if (r > t->maxrows)
    tableau_set_max_rows (t,r);
  if (c > t->maxcolumns)
    tableau_set_max_columns (t,c);
  
  t->columns = c;
  t->rows = r;
  init_tableau (t);
}


void
set_tableau_for_problem (struct TABLEAU *t, struct LP_PROBLEM *lp)
{
  int i,j;
  Real *rp;
  
  t->no_eq = lp->no_eq_constraints;
  t->no_cons = lp->no_constraints;
  t->no_ineq = t->no_cons - t->no_eq;
  t->no_vars = lp->no_vars;
  
  tableau_set_rowscolumns (t,
			   t->no_cons + 1,
			   t->no_ineq + t->no_vars + 1);

  /* Generate initial tableau. */
  for (rp = t->elements[0], j = 0; j < t->no_vars; j++)
    rp[j] = -lp->c[j];

  for(i = 0; i < t->no_cons; i++)
    {
      for(j = 0; j < t->no_vars; j++)
	t->elements[i+1][j] = lp->A[i][j];
    }

  for(j = 0; j < t->no_cons; j++)
    t->elements[j+1][t->columns-1] = lp->b[j];
  

  for(i = 0; i < t->no_ineq; i++)
    t->elements[i+1][lp->no_vars+i] = 1.0;

  for(i = 1; i < t->no_cons+1; i++)
    t->chk[i] = -1;	/* Arrays_P->Indexing problem; Algol
			 * version treates 0 as out
			 * of bounds, in C we prefer -1.
			 */

  if(debug_minit)
   print_tableau(t); /* Print the tableau. */


}

void
tableau_set_max_rows (struct TABLEAU*t, int maxrows)
{
  int lastmax = t->maxrows;
  int i;
  
  t->maxrows = maxrows;
  t->jmax =  xrealloc(t->jmax, maxrows*sizeof(int));
  t->neg_rhs =  xrealloc( t->neg_rhs, maxrows*sizeof(int));
  t->chk =  xrealloc( t->chk, maxrows*sizeof(int));
  t->delmax =  xrealloc(t->delmax,maxrows*sizeof(Real));

  /* if making smaller: junk excess rows */
  for (i = maxrows; i < lastmax; i++)
    free (t->elements[i]);
  
  t->elements= xrealloc(t->elements,maxrows*sizeof(Real *));

  /* if making larger: alloc new rows. */
  for(i = lastmax; i < maxrows; i++)
    t->elements[i]= xcalloc(t->maxcolumns, sizeof(Real));
}

void tableau_set_max_columns (struct TABLEAU *t, int maxcolumns)
{
  int i;

  t->maxcolumns = maxcolumns;

  t->neg_costcoef_indices =  xrealloc(t->neg_costcoef_indices, maxcolumns*sizeof(int));
  t->imin =  xrealloc(t->imin, maxcolumns*sizeof(int));
  t->thmin =  xrealloc(t->thmin, maxcolumns*sizeof(Real));

  for(i = 0; i < t->maxrows;  i++)
    t->elements[i]= xrealloc(t->elements[i], maxcolumns*sizeof(Real));
}

void init_tableau (struct TABLEAU *t)
{
  int i;
  int r = t->rows;
  int c = t->columns;
  
  memset (t->jmax, 0, sizeof (int) * r);
  memset (t->neg_rhs, 0, sizeof (int) * r);
  memset (t->chk, 0, sizeof (int) * r);
  memset (t->delmax, 0, sizeof (Real) * r);

  for(i = 0; i < r; i++)
      memset (t->elements[i], 0, sizeof (Real) * c);

  memset (t->neg_costcoef_indices, 0, sizeof (int) * c);
  memset (t->imin, 0, sizeof (int) * c);
  memset (t->thmin, 0, sizeof (Real) * c);  
}


void print_tableau(struct TABLEAU *t)
{
  register int i, j;
  
  printf ("TABLEAU (%d x %d, max %d x %d):\n", t->rows, t->columns, t->maxrows, t->maxcolumns);
  for(i = 0; i < t->rows; i++)
    {
      for(j = 0; j < t->columns; j++)
	printf("%6.3f ",t->elements[i][j]);

      printf("\n");
    }
}


void
tableau_read_solution (struct TABLEAU *tabp, struct LP_SOLUTION *sol)
{
  int i,j;
  
 init_solution (sol);
 sol->val = tabp->elements[0][tabp->columns-1];

 for(i = 1; i < tabp->rows; i++)
   {
     if(tabp->chk[i] >= sol->n)
       tabp->chk[i] = -1;

     if(tabp->chk[i] > -1)
       sol->prim[tabp->chk[i]] = tabp->elements[i][tabp->columns-1];
   }

 for(j = sol->n; j < tabp->columns - 1; j++)
   sol->dual[j-sol->n] = tabp->elements[0][j];
}
